import asyncio
import base64
from datetime import datetime

import aiohttp
import pytest
from aiohttp import web
from asn1crypto import algos
from certomancer.registry import CertLabel
from csc_dummy.csc_dummy_server import CSCWithCertomancer, DummyServiceParams
from freezegun import freeze_time
from pyhanko.sign.general import SigningError
from pyhanko.sign.signers import csc_signer
from pyhanko.sign.validation.utils import validate_raw

from .csc_utils.csc_dummy_client import CSCDummyClientAuthManager
from .samples import CERTOMANCER, TESTING_CA

SIGNER_B64 = """
MIIEMDCCAxigAwIBAgICEAEwDQYJKoZIhvcNAQELBQAwWTELMAkGA1UEBhMCQkUxFDASBgNVBAoM
C0V4YW1wbGUgSW5jMRowGAYDVQQLDBFUZXN0aW5nIEF1dGhvcml0eTEYMBYGA1UEAwwPSW50ZXJt
ZWRpYXRlIENBMCIYDzIwMjAwMTAxMDAwMDAwWhgPMjAyMjAxMDEwMDAwMDBaMHExCzAJBgNVBAYT
AkJFMRQwEgYDVQQKDAtFeGFtcGxlIEluYzEaMBgGA1UECwwRVGVzdGluZyBBdXRob3JpdHkxDjAM
BgNVBAMMBUFsaWNlMSAwHgYJKoZIhvcNAQkBFhFhbGljZUBleGFtcGxlLmNvbTCCASIwDQYJKoZI
hvcNAQEBBQADggEPADCCAQoCggEBAOFzgm4eL34uvUYrX4akyEBi+sn0gCYo8UOthApfluxF4cca
GhCHdjZa1PwRpV3bDGFQpUbhNu0juCBkYbRGxasQOn1CUDF7DCCjztNEd779WwRlA5dnqWMFU5Ij
toavSYl+CA1Ase2edxq7UjEZr4kIm7ADlUVpdKxLItJFEP4QOjqv5sENuiGCKpMqb/JGmvnLxRev
xDZQ8hIDV2s07krCog8hRChie39mDNmZ/RH/JbgME6mGY99bDAnhu8xH41iBo8GemEmmFesx8YPr
MivcXHk3QNt2LsKCAGZlG51fsrtiC31732W0+dc09PoITS0NMvP8/38dQmod3ktJBusCAwEAAaOB
5TCB4jAdBgNVHQ4EFgQUXsenutQKZ62drno9rRRqOLVzCrgwHwYDVR0jBBgwFoAU7796UYsupC5K
XdLO52F4zmo2SU8wDgYDVR0PAQH/BAQDAgbAMEcGA1UdHwRAMD4wPKA6oDiGNmh0dHA6Ly9weWhh
bmtvLnRlc3RzL3Rlc3RpbmctY2EvY3Jscy9pbnRlcm0vbGF0ZXN0LmNybDBHBggrBgEFBQcBAQQ7
MDkwNwYIKwYBBQUHMAGGK2h0dHA6Ly9weWhhbmtvLnRlc3RzL3Rlc3RpbmctY2Evb2NzcC9pbnRl
cm0wDQYJKoZIhvcNAQELBQADggEBACwIBhziVaZdlWg5S7PCjnL2yEeDnJWW3c1DXzKz8++mjDR0
GNsMgy1+XxKBR7gqaIDCUIPvgWko6UNfvt74txy1eBI4KHfWG/J3R9S54MTm+jZc/ctR9ma2kqcp
nz5p6HjfgdN/ejY3Fop+dZSvapbQumk27/3boKjzftXJP1VD6LSpINBDEX4kw9w4xZDpgZOcxB+s
k36fJ/cT50B7dYFFfSfKWCU7cpOYtkwnytfOBc4PxodIjrv1Y5ZSELyC8mY+U3zTlDLdG364BMSV
x4AzyPTjm252vITdbnsFZTPkpEz2gfQnh0Ee5jq+vkPfUyB0tgGju0yu9EgFIu2pzGY=
"""

INTERM_B64 = """
MIID2TCCAsGgAwIBAgICEAEwDQYJKoZIhvcNAQELBQAwUTELMAkGA1UEBhMCQkUxFDASBgNVBAoM
C0V4YW1wbGUgSW5jMRowGAYDVQQLDBFUZXN0aW5nIEF1dGhvcml0eTEQMA4GA1UEAwwHUm9vdCBD
QTAiGA8yMDAwMDEwMTAwMDAwMFoYDzIxMDAwMTAxMDAwMDAwWjBZMQswCQYDVQQGEwJCRTEUMBIG
A1UECgwLRXhhbXBsZSBJbmMxGjAYBgNVBAsMEVRlc3RpbmcgQXV0aG9yaXR5MRgwFgYDVQQDDA9J
bnRlcm1lZGlhdGUgQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDB6drP3ogV8UnY
2YM0GREPUCCEe7FX905/Ocl2MjyPyh3/vaf542LGQWdC1mf3fNQsMu2dqaZTwUQrKJwIs07QV7r1
now3rKUNG8YyGtdZfKyt6jgOsfaZ5KB54sUtWGy4LKT4cPq0vEWnm8lFo91kvJqVOjFXWpUSKwTP
YJV+riBmKg25C+8D2L5oKdMzQxd8o9K0kAI0fDF1CPbqtCB/GIzwN2Cq20bX9dpzTbZOlvIundA4
hLYBNcfXnigPLylc3x3YxypoBotQ5L9MxKceBgrv1SfgtEAqxeYgUZe/UQdEG7jRy0JIYFKS/EWv
HcJ6jAX+v64hKeIAslO/67TbAgMBAAGjga4wgaswHQYDVR0OBBYEFO+/elGLLqQuSl3SzudheM5q
NklPMB8GA1UdIwQYMBaAFPG0eZUCsrRfOXL9sI/CX69bmlC1MBIGA1UdEwEB/wQIMAYBAf8CAQAw
DgYDVR0PAQH/BAQDAgGGMEUGA1UdHwQ+MDwwOqA4oDaGNGh0dHA6Ly9weWhhbmtvLnRlc3RzL3Rl
c3RpbmctY2EvY3Jscy9yb290L2xhdGVzdC5jcmwwDQYJKoZIhvcNAQELBQADggEBAAx7qZLSOIk4
B7w6ymmPF3cKrafdJg83NhWAJZHOie9E23hhOV0uOLoaab2m+TbpqUXXpeuxCC3t2duLueSceHUX
Y+UhZGF4LdTojK5yVnyHynJVwbAAzRyO1hIjI9SFW+NMsFNwN6t50RAhTGY7jZvvv2y3mMwUHE2u
lgsQNax0Tyoh6yOtb752IWm0Yj+cEMQ36cjxkcFZgYTiziuYkYJM8wL+kADG+HUsnfdYJWMM/lJ7
YEkSIm0lhec3xR6siySnZI2whU3twY37WcnxwjkAczSdHZJpyzrSyjp8rCwk0zOF4HGQr3WuJhHo
kS0gHz5+cFMDBZxeP5Oq5UFvFtU=
"""


def test_proc_cert_info_response():
    cred_info = csc_signer._process_certificate_info_response(
        {
            'cert': {'certificates': [SIGNER_B64, INTERM_B64]},
            'key': {
                'algo': ['1.2.840.113549.1.1.11'],
                'len': 2048,
            },
            'multisign': 10,
            'authMode': 'implicit',
            'SCAL': 2,
        }
    )
    assert cred_info.max_batch_size == 10
    assert cred_info.hash_pinning_required
    assert 'sha256_rsa' in cred_info.supported_mechanisms
    assert 'Alice' in cred_info.signing_cert.subject.human_friendly


@pytest.mark.parametrize(
    'response_data,err_msg',
    [
        (
            {
                'cert': {},
                'key': {
                    'algo': ['1.2.840.113549.1.1.11'],
                    'len': 2048,
                },
                'multisign': 10,
                'authMode': 'implicit',
                'SCAL': 2,
            },
            "Could not retrieve certificates from response",
        ),
        (
            {
                'key': {
                    'algo': ['1.2.840.113549.1.1.11'],
                    'len': 2048,
                },
                'multisign': 10,
                'authMode': 'implicit',
                'SCAL': 2,
            },
            "Could not retrieve certificates from response",
        ),
        (
            {
                'cert': {'certificates': [SIGNER_B64.replace('M', 'Z')]},
                'key': {
                    'algo': ['1.2.840.113549.1.1.11'],
                    'len': 2048,
                },
                'multisign': 10,
                'authMode': 'implicit',
                'SCAL': 2,
            },
            "Could not decode certificates in response",
        ),
        (
            {
                'cert': {'certificates': [SIGNER_B64, INTERM_B64]},
                'key': {
                    'algo': '1.2.840.113549.1.1.11',
                    'len': 2048,
                },
                'multisign': 10,
                'authMode': 'implicit',
                'SCAL': 2,
            },
            "Could not retrieve supported signing mechanisms",
        ),
        (
            {
                'cert': {'certificates': [SIGNER_B64, INTERM_B64]},
                'key': {
                    'algo': ['zzz1.2.840.113549.1.1.11'],
                    'len': 2048,
                },
                'multisign': 10,
                'authMode': 'implicit',
                'SCAL': 2,
            },
            "Could not retrieve supported signing mechanisms",
        ),
        (
            {
                'cert': {'certificates': [SIGNER_B64, INTERM_B64]},
                'key': {
                    'len': 2048,
                },
                'multisign': 10,
                'authMode': 'implicit',
                'SCAL': 2,
            },
            "Could not retrieve supported signing mechanisms",
        ),
        (
            {
                'cert': {'certificates': [SIGNER_B64, INTERM_B64]},
                'key': {
                    'algo': ['1.2.840.113549.1.1.11'],
                    'len': 2048,
                },
                'authMode': 'implicit',
                'SCAL': 2,
            },
            "Could not retrieve max batch size",
        ),
        (
            {
                'cert': {'certificates': [SIGNER_B64, INTERM_B64]},
                'key': {
                    'algo': ['1.2.840.113549.1.1.11'],
                    'len': 2048,
                },
                'multisign': 'foobar',
                'authMode': 'implicit',
                'SCAL': 2,
            },
            "Could not retrieve max batch size",
        ),
        (
            {
                'cert': {'certificates': [SIGNER_B64, INTERM_B64]},
                'key': {
                    'algo': ['1.2.840.113549.1.1.11'],
                    'len': 2048,
                },
                'multisign': 10,
                'authMode': 'implicit',
                'SCAL': 3,
            },
            "SCAL value must be",
        ),
        (
            {
                'cert': {'certificates': [SIGNER_B64, INTERM_B64]},
                'key': {
                    'algo': ['1.2.840.113549.1.1.11'],
                    'len': 2048,
                },
                'multisign': 10,
                'authMode': 'implicit',
                'SCAL': "zzz",
            },
            "SCAL value must be",
        ),
    ],
)
def test_proc_cert_info_response_errors(response_data, err_msg):
    with pytest.raises(SigningError, match=err_msg):
        csc_signer._process_certificate_info_response(response_data)


def test_format_csc_auth_request():
    # any old auth manager will do
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=csc_signer.CSCServiceSessionInfo(
            'https://example.com', 'foobar'
        ),
        credential_info=csc_signer.CSCCredentialInfo(
            signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
            chain=[],
            supported_mechanisms=frozenset(),
            max_batch_size=10,
            hash_pinning_required=False,
            response_data={},
        ),
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=''),
    )
    result = auth_man.format_csc_auth_request(
        pin='1234',
        otp='123456',
        hash_b64s=[
            'Sa6Tcy/PjWP+HM51lmSYLb1bIxYfAH26hWGGKtyW0GM=',
            'Sa6Tcy/PjWP+HM51lmSYLb1bIxYfAH26hWGGKtyW0GM=',
        ],
        description='baz',
        client_data='quux',
    )
    assert result == {
        'credentialID': 'foobar',
        'numSignatures': 2,
        'hash': [
            'Sa6Tcy/PjWP+HM51lmSYLb1bIxYfAH26hWGGKtyW0GM=',
            'Sa6Tcy/PjWP+HM51lmSYLb1bIxYfAH26hWGGKtyW0GM=',
        ],
        'PIN': '1234',
        'OTP': '123456',
        'description': 'baz',
        'clientData': 'quux',
    }


@freeze_time('2021-11-01T00:00:00+00:00')
def test_parse_csc_auth_response():
    response_data = {'SAD': 'foobar', 'expiresIn': 300}

    expires_at = datetime.fromisoformat('2021-11-01T00:05:00+00:00')

    result = csc_signer.CSCAuthorizationManager.parse_csc_auth_response(
        response_data
    )
    assert result == csc_signer.CSCAuthorizationInfo(
        sad='foobar', expires_at=expires_at
    )


@freeze_time('2021-11-01T00:00:00+00:00')
def test_parse_csc_auth_response_default_expiry():
    response_data = {'SAD': 'foobar'}

    expires_at = datetime.fromisoformat('2021-11-01T01:00:00+00:00')

    result = csc_signer.CSCAuthorizationManager.parse_csc_auth_response(
        response_data
    )
    assert result == csc_signer.CSCAuthorizationInfo(
        sad='foobar', expires_at=expires_at
    )


@pytest.mark.parametrize(
    'response_data,err_msg',
    [
        ({}, "Could not extract SAD"),
        (
            {'SAD': 'foobar', 'expiresIn': 'nonsense'},
            "Could not process expiresIn",
        ),
    ],
)
def test_parse_csc_auth_response_error(response_data, err_msg):
    with pytest.raises(SigningError, match=err_msg):
        csc_signer.CSCAuthorizationManager.parse_csc_auth_response(
            response_data
        )


# network failure tests
@pytest.mark.asyncio
async def test_cert_provision_fail():
    csc_session_info = csc_signer.CSCServiceSessionInfo(
        'https://example.invalid', 'foobar'
    )
    with pytest.raises(SigningError, match='Credential info request failed'):
        async with aiohttp.ClientSession() as session:
            await csc_signer.fetch_certs_in_csc_credential(
                session, csc_session_info=csc_session_info
            )


@pytest.mark.asyncio
async def test_sign_mechanism_not_supported():
    csc_session_info = csc_signer.CSCServiceSessionInfo(
        'https://example.com', 'foobar'
    )
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=csc_session_info,
        credential_info=csc_signer.CSCCredentialInfo(
            signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
            chain=[],
            supported_mechanisms=frozenset({'is_nonsense'}),
            max_batch_size=10,
            hash_pinning_required=False,
            response_data={},
        ),
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=''),
    )
    # check expected failure for a signing attempt
    with pytest.raises(SigningError, match='No signing results available'):
        async with aiohttp.ClientSession() as session:
            signer = csc_signer.CSCSigner(session, auth_manager=auth_man)
            await signer.async_sign_raw(b'foobarbazquux', 'sha256')

    # check expected failure when fetching the signature mechanism directly
    with pytest.raises(SigningError, match='must be one of'):
        # noinspection PyTypeChecker
        signer = csc_signer.CSCSigner(None, auth_manager=auth_man)
        signer.get_signature_mechanism_for_digest(digest_algorithm='sha256')

    # ...but overrides should still work
    # noinspection PyTypeChecker
    signer = csc_signer.CSCSigner(None, auth_manager=auth_man)
    signer._signature_mechanism = mech = algos.SignedDigestAlgorithm(
        {'algorithm': 'sha256_rsa'}
    )
    assert (
        signer.get_signature_mechanism_for_digest(digest_algorithm='sha256')
        == mech
    )


@pytest.mark.asyncio
async def test_sign_network_fail():
    csc_session_info = csc_signer.CSCServiceSessionInfo(
        'https://example.invalid', 'foobar'
    )
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=csc_session_info,
        credential_info=csc_signer.CSCCredentialInfo(
            signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
            chain=[],
            supported_mechanisms=frozenset({'sha256_rsa'}),
            max_batch_size=10,
            hash_pinning_required=False,
            response_data={},
        ),
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=''),
    )

    with pytest.raises(SigningError, match='No signing results available'):
        async with aiohttp.ClientSession() as session:
            signer = csc_signer.CSCSigner(session, auth_manager=auth_man)
            await signer.async_sign_raw(b'foobarbazquux', 'sha256')


@pytest.mark.asyncio
async def test_sign_wrong_number_of_sigs(aiohttp_client):
    csc_session_info = csc_signer.CSCServiceSessionInfo('', 'foobar')
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=csc_session_info,
        credential_info=csc_signer.CSCCredentialInfo(
            signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
            chain=[],
            supported_mechanisms=frozenset({'sha256_rsa'}),
            max_batch_size=2,
            hash_pinning_required=False,
            response_data={},
        ),
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=''),
    )

    async def fake_return(_request):
        return web.json_response(
            {
                'signatures': [
                    base64.b64encode(bytes(512)).decode('ascii'),
                ]
            }
        )

    app = web.Application()
    app.router.add_post('/csc/v1/signatures/signHash', fake_return)
    client = await aiohttp_client(app)
    # noinspection PyTypeChecker
    signer = csc_signer.CSCSigner(
        client, auth_manager=auth_man, batch_size=2, batch_autocommit=False
    )
    result = asyncio.gather(
        signer.async_sign_raw(b'foobarbaz', 'sha256'),
        signer.async_sign_raw(b'foobarbazquux', 'sha256'),
    )

    with pytest.raises(SigningError, match='Expected 2 signatures'):
        await asyncio.sleep(1)
        await signer.commit()
    try:
        result.cancel()
        await result
    except asyncio.CancelledError:
        pass


@pytest.mark.parametrize('response_obj', [{'signatures': [None]}, {}])
@pytest.mark.asyncio
async def test_sign_unreadable_sig(aiohttp_client, response_obj):
    csc_session_info = csc_signer.CSCServiceSessionInfo('', 'foobar')
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=csc_session_info,
        credential_info=csc_signer.CSCCredentialInfo(
            signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
            chain=[],
            supported_mechanisms=frozenset({'sha256_rsa'}),
            max_batch_size=1,
            hash_pinning_required=False,
            response_data={},
        ),
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=''),
    )

    async def fake_return(_request):
        return web.json_response(response_obj)

    app = web.Application()
    app.router.add_post('/csc/v1/signatures/signHash', fake_return)
    client = await aiohttp_client(app)
    # noinspection PyTypeChecker
    signer = csc_signer.CSCSigner(
        client,
        auth_manager=auth_man,
        batch_size=1,
        batch_autocommit=False,
        client_data='Some client data, because why not',
    )
    result = asyncio.create_task(signer.async_sign_raw(b'foobarbaz', 'sha256'))
    with pytest.raises(SigningError, match='Expected response with b64'):
        await asyncio.sleep(1)
        await signer.commit()

    try:
        result.cancel()
        await result
    except asyncio.CancelledError:
        pass


@pytest.mark.asyncio
async def test_fail_different_digest():
    csc_session_info = csc_signer.CSCServiceSessionInfo('', 'foobar')
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=csc_session_info,
        credential_info=csc_signer.CSCCredentialInfo(
            signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
            chain=[],
            supported_mechanisms=frozenset({'sha256_rsa'}),
            max_batch_size=2,
            hash_pinning_required=False,
            response_data={},
        ),
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=''),
    )
    # noinspection PyTypeChecker
    signer = csc_signer.CSCSigner(None, auth_manager=auth_man, batch_size=2)
    with pytest.raises(SigningError, match='same digest function'):
        result = asyncio.gather(
            signer.async_sign_raw(b'foobarbaz', 'sha256'),
            signer.async_sign_raw(b'foobarbazquux', 'sha512'),
        )
        await result


async def _set_up_dummy_client(aiohttp_client, require_hash_pinning=True):
    csc_session_info = csc_signer.CSCServiceSessionInfo(
        '', 'testing-ca/signer1'
    )
    csc_dummy = CSCWithCertomancer(
        certomancer_config=CERTOMANCER,
        service_params=DummyServiceParams(
            hash_pinning_required=require_hash_pinning
        ),
    )
    csc_dummy.register_routes()
    client = await aiohttp_client(csc_dummy.app)

    auth_man = CSCDummyClientAuthManager(
        client,
        session_info=csc_session_info,
        credential_info=await csc_signer.fetch_certs_in_csc_credential(
            client, csc_session_info
        ),
    )
    return client, auth_man, csc_dummy


@pytest.mark.asyncio
async def test_submit_job_during_commit(aiohttp_client):
    client, auth_man, csc_dummy = await _set_up_dummy_client(aiohttp_client)

    class SlowCommitter(csc_signer.CSCSigner):
        _committed_once = False

        async def _do_commit(self, batch):
            # waste time
            if not self._committed_once:
                await asyncio.sleep(5)
            self._committed_once = True
            await super()._do_commit(batch)

    signer = SlowCommitter(
        session=client,
        auth_manager=auth_man,
        batch_autocommit=True,
        batch_size=2,
    )

    async def make_first_sig():
        # submit an incomplete batch
        result = asyncio.create_task(
            signer.async_sign_raw(b'foobar1', 'sha256'),
        )
        await asyncio.sleep(1)
        await signer.commit()
        return await result

    async def make_other_sigs():
        await asyncio.sleep(2)
        return await asyncio.gather(
            signer.async_sign_raw(b'foobar2', 'sha256'),
            signer.async_sign_raw(b'foobar3', 'sha256'),
        )

    sig1, others = await asyncio.gather(make_first_sig(), make_other_sigs())

    signer_cert = TESTING_CA.get_cert(CertLabel('signer1'))
    for ix, sig in enumerate([sig1, *others]):
        validate_raw(
            sig,
            b'foobar%d' % (ix + 1),
            signer_cert,
            signature_algorithm=algos.SignedDigestAlgorithm(
                {'algorithm': 'sha256_rsa'}
            ),
            md_algorithm='sha256',
        )
    assert auth_man.authorizations_requested == 2


@pytest.mark.asyncio
async def test_multi_commit_failure(aiohttp_client):
    client, auth_man, csc_dummy = await _set_up_dummy_client(aiohttp_client)
    # deliberately pass a bogus SAD to make the commit fail
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=auth_man.csc_session_info,
        credential_info=auth_man.credential_info,
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=''),
    )

    class SlowCommitter(csc_signer.CSCSigner):
        async def _do_commit(self, batch):
            # waste time
            await asyncio.sleep(3)
            await super()._do_commit(batch)

    signer = SlowCommitter(
        session=client,
        auth_manager=auth_man,
        batch_autocommit=False,
        batch_size=1,
    )

    async def produce_signature():
        result = asyncio.create_task(
            signer.async_sign_raw(b'foobar', 'sha256'),
        )
        await asyncio.sleep(1)
        with pytest.raises(SigningError, match='Signature request failed'):
            await signer.commit()
        with pytest.raises(SigningError, match='No signing results'):
            return await result

    async def commit_again():
        with pytest.raises(SigningError, match='Commit failed'):
            await asyncio.sleep(2)
            await signer.commit()

    await asyncio.gather(produce_signature(), commit_again())
    # this should now return immediately as there is no batch
    await signer.commit()


@pytest.mark.asyncio
async def test_csc_with_parameters(aiohttp_client):
    #  produce a signature with parameters

    client, auth_man, csc_dummy = await _set_up_dummy_client(aiohttp_client)

    signer = csc_signer.CSCSigner(
        session=client,
        auth_manager=auth_man,
        batch_autocommit=True,
        batch_size=1,
        prefer_pss=True,
    )

    result = await signer.async_sign_raw(b'foobar', digest_algorithm='sha256')
    signer_cert = TESTING_CA.get_cert(CertLabel('signer1'))
    mech = signer.get_signature_mechanism_for_digest('sha256')
    assert mech.signature_algo == 'rsassa_pss'
    validate_raw(
        result,
        b'foobar',
        signer_cert,
        signature_algorithm=mech,
        md_algorithm='sha256',
    )


@pytest.mark.asyncio
async def test_prefetched_sad_not_twice(aiohttp_client):
    client, auth_man, csc_dummy = await _set_up_dummy_client(
        aiohttp_client, require_hash_pinning=False
    )

    # prefetch SAD that is not bound to any hashes
    async with client.post(
        '/csc/v1/credentials/authorize',
        json=auth_man.format_csc_auth_request(),
        raise_for_status=True,
    ) as resp:
        sad = (await resp.json())['SAD']
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=auth_man.csc_session_info,
        credential_info=auth_man.credential_info,
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=sad),
    )

    signer = csc_signer.CSCSigner(
        session=client,
        auth_manager=auth_man,
        batch_autocommit=True,
        batch_size=1,
    )

    result = await signer.async_sign_raw(b'foobar', digest_algorithm='sha256')
    signer_cert = TESTING_CA.get_cert(CertLabel('signer1'))
    validate_raw(
        result,
        b'foobar',
        signer_cert,
        signature_algorithm=algos.SignedDigestAlgorithm(
            {'algorithm': 'sha256_rsa'}
        ),
        md_algorithm='sha256',
    )

    # but a second attempt should fail
    with pytest.raises(SigningError, match='No signing results'):
        await signer.async_sign_raw(b'foobar', digest_algorithm='sha256')


# The API docs say that the placeholder will be 512 bytes, so we record
# that in a test here


@pytest.mark.asyncio
async def test_csc_placeholder_sig_size():
    csc_session_info = csc_signer.CSCServiceSessionInfo(
        'https://example.com', 'foobar'
    )
    auth_man = csc_signer.PrefetchedSADAuthorizationManager(
        csc_session_info=csc_session_info,
        credential_info=csc_signer.CSCCredentialInfo(
            signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
            chain=[],
            supported_mechanisms=frozenset({'is_nonsense'}),
            max_batch_size=10,
            hash_pinning_required=False,
            response_data={},
        ),
        csc_auth_info=csc_signer.CSCAuthorizationInfo(sad=''),
    )
    # noinspection PyTypeChecker
    signer = csc_signer.CSCSigner(None, auth_manager=auth_man)
    await signer.async_sign_raw(b'foobarbazquux', 'sha256', dry_run=True)



from datetime import datetime
from typing import Optional

import tzlocal
from asn1crypto import algos, cms, core, keys, tsp, x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from pyhanko_certvalidator.registry import CertificateStore

from .. import general
from ..general import get_pyca_cryptography_hash, simple_cms_attribute
from .api import TimeStamper
from .common_utils import get_nonce

__all__ = ['DummyTimeStamper']


class DummyTimeStamper(TimeStamper):
    """
    Timestamper that acts as its own TSA. It accepts all requests and
    signs them using the certificate provided.
    Used for testing purposes.
    """

    def __init__(
        self,
        tsa_cert: x509.Certificate,
        tsa_key: keys.PrivateKeyInfo,
        certs_to_embed: Optional[CertificateStore] = None,
        fixed_dt: Optional[datetime] = None,
        include_nonce=True,
        override_md=None,
    ):
        self.tsa_cert = tsa_cert
        self.tsa_key = tsa_key
        self.certs_to_embed = list(certs_to_embed or ())
        self.fixed_dt = fixed_dt
        self.override_md = override_md
        super().__init__(include_nonce=include_nonce)

    def request_tsa_response(self, req: tsp.TimeStampReq) -> tsp.TimeStampResp:
        # We pretend that certReq is always true in the request

        # TODO generalise my detached signature logic to include cases like this
        #  (see § 5.4 in RFC 5652)
        # TODO does the RFC
        status = tsp.PKIStatusInfo({'status': tsp.PKIStatus('granted')})
        message_imprint: tsp.MessageImprint = req['message_imprint']
        md_algorithm = self.override_md
        if md_algorithm is None:
            md_algorithm = message_imprint['hash_algorithm']['algorithm'].native
        digest_algorithm_obj = algos.DigestAlgorithm(
            {'algorithm': md_algorithm}
        )
        dt = self.fixed_dt or datetime.now(tz=tzlocal.get_localzone())
        tst_info_args = {
            'version': 'v1',
            # See http://oidref.com/1.3.6.1.4.1.4146.2.2
            # I don't really care too much, this is a testing device anyway
            'policy': tsp.ObjectIdentifier('1.3.6.1.4.1.4146.2.2'),
            'message_imprint': message_imprint,
            # should be sufficiently random (again, this is a testing class)
            'serial_number': get_nonce(),
            'gen_time': dt,
            'tsa': x509.GeneralName(
                name='directory_name', value=self.tsa_cert.subject
            ),
        }
        if req['nonce'].native is not None:
            tst_info_args['nonce'] = req['nonce']

        tst_info = tsp.TSTInfo(tst_info_args)
        tst_info_data = tst_info.dump()
        md_spec = get_pyca_cryptography_hash(md_algorithm)
        md = hashes.Hash(md_spec)
        md.update(tst_info_data)
        message_digest_value = md.finalize()
        signed_attrs = cms.CMSAttributes(
            [
                simple_cms_attribute('content_type', 'tst_info'),
                simple_cms_attribute(
                    'signing_time', cms.Time({'utc_time': core.UTCTime(dt)})
                ),
                simple_cms_attribute(
                    'signing_certificate',
                    general.as_signing_certificate(self.tsa_cert),
                ),
                simple_cms_attribute('message_digest', message_digest_value),
            ]
        )
        priv_key = serialization.load_der_private_key(
            self.tsa_key.dump(), password=None
        )
        if not isinstance(priv_key, RSAPrivateKey):
            raise NotImplementedError("Dummy timestamper is RSA-only.")
        signature = priv_key.sign(
            signed_attrs.dump(),
            PKCS1v15(),
            get_pyca_cryptography_hash(md_algorithm.upper()),
        )
        sig_info = cms.SignerInfo(
            {
                'version': 'v1',
                'sid': cms.SignerIdentifier(
                    {
                        'issuer_and_serial_number': cms.IssuerAndSerialNumber(
                            {
                                'issuer': self.tsa_cert.issuer,
                                'serial_number': self.tsa_cert.serial_number,
                            }
                        )
                    }
                ),
                'digest_algorithm': digest_algorithm_obj,
                'signature_algorithm': algos.SignedDigestAlgorithm(
                    {'algorithm': 'rsassa_pkcs1v15'}
                ),
                'signed_attrs': signed_attrs,
                'signature': signature,
            }
        )
        certs = set(self.certs_to_embed)
        certs.add(self.tsa_cert)
        signed_data = {
            # must use v3 to get access to the EncapsulatedContentInfo construct
            'version': 'v3',
            'digest_algorithms': cms.DigestAlgorithms((digest_algorithm_obj,)),
            'encap_content_info': cms.EncapsulatedContentInfo(
                {
                    'content_type': cms.ContentType('tst_info'),
                    'content': cms.ParsableOctetString(tst_info_data),
                }
            ),
            'certificates': certs,
            'signer_infos': [sig_info],
        }
        tst = cms.ContentInfo(
            {
                'content_type': cms.ContentType('signed_data'),
                'content': cms.SignedData(signed_data),
            }
        )
        return tsp.TimeStampResp({'status': status, 'time_stamp_token': tst})

    async def async_request_tsa_response(
        self, req: tsp.TimeStampReq
    ) -> tsp.TimeStampResp:
        # just do it synchronously
        return self.request_tsa_response(req)



from typing import List, Optional

import click
from pyhanko import __version__
from pyhanko.cli._root import cli_root
from pyhanko.cli._trust import (
    _get_key_usage_settings,
    build_vc_kwargs,
    trust_options,
)
from pyhanko.cli.commands.signing.plugin import command_from_plugin
from pyhanko.cli.commands.stamp import select_style
from pyhanko.cli.utils import parse_field_location_spec
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.sign import DEFAULT_SIGNER_KEY_USAGE, fields, signers
from pyhanko.sign.signers.pdf_byterange import BuildProps
from pyhanko.sign.timestamps import HTTPTimeStamper

from pyhanko_certvalidator import ValidationContext

from ..._ctx import CLIContext
from ...plugin_api import SigningCommandPlugin

__all__ = ['signing', 'addsig', 'register']

from ...runtime import pyhanko_exception_manager


@cli_root.group(help='sign PDFs and other files', name='sign')
def signing():
    pass


@trust_options
@signing.group(name='addsig', help='add a signature')
@click.option(
    '--field',
    help=(
        'signature field name, or field specification PAGE/X1,Y1,X2,Y2/NAME '
        '(required unless the field contains exactly one signature field)'
    ),
    required=False,
)
@click.option('--name', help='explicitly specify signer name', required=False)
@click.option('--reason', help='reason for signing', required=False)
@click.option('--location', help='location of signing', required=False)
@click.option('--contact-info', help='contact of the signer', required=False)
@click.option(
    '--certify',
    help='add certification signature',
    required=False,
    default=False,
    is_flag=True,
    type=bool,
    show_default=True,
)
@click.option(
    '--existing-only',
    help='never create signature fields',
    required=False,
    default=False,
    is_flag=True,
    type=bool,
    show_default=True,
)
@click.option(
    '--timestamp-url',
    help='URL for timestamp server',
    required=False,
    type=str,
    default=None,
)
@click.option(
    '--use-pades',
    help='sign PAdES-style [level B/B-T/B-LT/B-LTA]',
    required=False,
    default=False,
    is_flag=True,
    type=bool,
    show_default=True,
)
@click.option(
    '--use-pades-lta',
    help='produce PAdES-B-LTA signature',
    required=False,
    default=False,
    is_flag=True,
    type=bool,
    show_default=True,
)
@click.option(
    '--prefer-pss',
    is_flag=True,
    default=False,
    type=bool,
    help='prefer RSASSA-PSS to PKCS#1 v1.5 padding, if available',
)
@click.option(
    '--with-validation-info',
    help='embed revocation info',
    required=False,
    default=False,
    is_flag=True,
    type=bool,
    show_default=True,
)
@click.option(
    '--style-name',
    help='stamp style name for signature appearance',
    required=False,
    type=str,
)
@click.option(
    '--stamp-url',
    help='QR code URL to use in QR stamp style',
    required=False,
    type=str,
)
@click.option(
    '--detach',
    type=bool,
    is_flag=True,
    default=False,
    help=(
        'write only the signature CMS object to the output file; '
        'this can be used to sign non-PDF files'
    ),
)
@click.option(
    '--detach-pem',
    help='output PEM data instead of DER when using --detach',
    type=bool,
    is_flag=True,
    default=False,
)
@click.option(
    '--retroactive-revinfo',
    help='Treat revocation info as retroactively valid '
    '(i.e. ignore thisUpdate timestamp)',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    '--no-strict-syntax',
    help='Attempt to ignore syntactical problems in the input file '
    'and enable signature creation in hybrid-reference files.'
    '(warning: such documents may behave in unexpected ways)',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.pass_context
def addsig(
    ctx: click.Context,
    field,
    name,
    reason,
    contact_info,
    location,
    certify,
    existing_only,
    timestamp_url,
    use_pades,
    use_pades_lta,
    with_validation_info,
    validation_context,
    trust_replace,
    trust,
    other_certs,
    style_name,
    stamp_url,
    prefer_pss,
    retroactive_revinfo,
    detach,
    detach_pem,
    no_strict_syntax,
):
    ctx_obj: CLIContext = ctx.obj
    ctx_obj.existing_fields_only = existing_only or field is None
    ctx_obj.timestamp_url = timestamp_url
    ctx_obj.prefer_pss = prefer_pss

    if detach or detach_pem:
        ctx_obj.detach_pem = detach_pem
        ctx_obj.sig_settings = None
        if field:
            raise click.ClickException(
                "--field is not compatible with --detach or --detach-pem"
            )
        return  # everything else doesn't apply

    if use_pades_lta:
        use_pades = with_validation_info = True
        if not timestamp_url:
            raise click.ClickException(
                "--timestamp-url is required for --use-pades-lta"
            )
    if use_pades:
        subfilter = fields.SigSeedSubFilter.PADES
    else:
        subfilter = fields.SigSeedSubFilter.ADOBE_PKCS7_DETACHED

    key_usage = DEFAULT_SIGNER_KEY_USAGE
    if with_validation_info:
        vc_kwargs = build_vc_kwargs(
            ctx.obj.config,
            validation_context,
            trust,
            trust_replace,
            other_certs,
            retroactive_revinfo,
            allow_fetching=True,
        )
        vc = ValidationContext(**vc_kwargs)
        key_usage_sett = _get_key_usage_settings(ctx, validation_context)
        if key_usage_sett is not None and key_usage_sett.key_usage is not None:
            key_usage = key_usage_sett.key_usage
    else:
        vc = None
    field_name: Optional[str]
    if field:
        field_name, new_field_spec = parse_field_location_spec(
            field, require_full_spec=False
        )
    else:
        field_name = new_field_spec = None
    if new_field_spec and existing_only:
        raise click.ClickException(
            "Specifying field coordinates is incompatible with --existing-only"
        )
    ctx_obj.sig_settings = signers.PdfSignatureMetadata(
        field_name=field_name,
        location=location,
        reason=reason,
        contact_info=contact_info,
        name=name,
        certify=certify,
        subfilter=subfilter,
        embed_validation_info=with_validation_info,
        validation_context=vc,
        signer_key_usage=key_usage,
        use_pades_lta=use_pades_lta,
        app_build_props=BuildProps(name='pyHanko CLI', revision=__version__),
    )
    ctx_obj.new_field_spec = new_field_spec
    ctx_obj.stamp_style = select_style(ctx, style_name, stamp_url)
    ctx_obj.stamp_url = stamp_url
    ctx_obj.lenient = no_strict_syntax
    ctx_obj.ux.visible_signature_desired = bool(style_name or new_field_spec)


def _unavailable(signer_plugin: SigningCommandPlugin):
    def _unavailable_callback():
        raise click.ClickException(
            signer_plugin.unavailable_message
            or "This subcommand is not available"
        )

    return _unavailable_callback


def register(plugins: List[SigningCommandPlugin]):
    # we reset the command list before (re)populating it, in order to
    # make the tests more consistent
    addsig.commands = {}
    for signer_plugin in plugins:
        if signer_plugin.is_available():
            addsig.add_command(command_from_plugin(signer_plugin))
        else:

            addsig.add_command(
                click.Command(
                    name=signer_plugin.subcommand_name,
                    help=signer_plugin.help_summary + " [unavailable]",
                    callback=_unavailable(signer_plugin),
                )
            )


readable_file = click.Path(exists=True, readable=True, dir_okay=False)
writable_file = click.Path(writable=True, dir_okay=False)


@trust_options
@signing.command(name='timestamp', help='add timestamp to PDF')
@click.argument('infile', type=readable_file)
@click.argument('outfile', type=writable_file)
@click.option(
    '--timestamp-url',
    help='URL for timestamp server',
    required=True,
    type=str,
    default=None,
)
@click.pass_context
def timestamp(
    ctx,
    infile,
    outfile,
    validation_context,
    trust,
    trust_replace,
    other_certs,
    timestamp_url,
):
    with pyhanko_exception_manager():
        vc_kwargs = build_vc_kwargs(
            ctx.obj.config,
            validation_context,
            trust,
            trust_replace,
            other_certs,
            retroactive_revinfo=True,
        )
        timestamper = HTTPTimeStamper(timestamp_url)
        with open(infile, 'rb') as inf:
            w = IncrementalPdfFileWriter(inf)
            pdf_timestamper = signers.PdfTimeStamper(timestamper)
            with open(outfile, 'wb') as outf:
                pdf_timestamper.timestamp_pdf(
                    w,
                    'sha256',
                    validation_context=ValidationContext(**vc_kwargs),
                    output=outf,
                )



import abc
from datetime import datetime
from typing import Iterable, Optional

from asn1crypto import algos, cms, core
from asn1crypto import pdf as asn1_pdf
from asn1crypto import tsp, x509
from cryptography.hazmat.primitives import hashes

from .general import (
    as_signing_certificate_v2,
    get_pyca_cryptography_hash,
    simple_cms_attribute,
)
from .timestamps import TimeStamper

__all__ = [
    'SignedAttributeProviderSpec',
    'UnsignedAttributeProviderSpec',
    'CMSAttributeProvider',
    'SigningTimeProvider',
    'SigningCertificateV2Provider',
    'AdobeRevinfoProvider',
    'CMSAlgorithmProtectionProvider',
    'TSTProvider',
]


class CMSAttributeProvider:
    """
    Base class to provide asynchronous CMS attribute values.
    """

    attribute_type: str
    """
    Name of the CMS attribute type this provider supplies. See
    :class:`cms.CMSAttributeType`.
    """

    async def build_attr_value(self, dry_run=False):
        """
        Build the attribute value asynchronously.

        :param dry_run:
            ``True`` if the signer is operating in dry-run (size estimation)
            mode.
        :return:
            An attribute value appropriate for the attribute type.
        """
        raise NotImplementedError

    async def get_attribute(self, dry_run=False) -> Optional[cms.CMSAttribute]:
        value = await self.build_attr_value(dry_run=dry_run)
        if value is not None:
            return simple_cms_attribute(self.attribute_type, value)
        else:
            return None


class SigningCertificateV2Provider(CMSAttributeProvider):
    """
    Provide a value for the signing-certificate-v2 attribute.

    :param signing_cert:
        Certificate containing the signer's public key.
    """

    attribute_type: str = 'signing_certificate_v2'

    def __init__(self, signing_cert: x509.Certificate):
        self.signing_cert = signing_cert

    async def build_attr_value(self, dry_run=False) -> tsp.SigningCertificateV2:
        return as_signing_certificate_v2(self.signing_cert)


class SigningTimeProvider(CMSAttributeProvider):
    """
    Provide a value for the signing-time attribute (i.e. an otherwise
    unauthenticated timestamp).

    :param timestamp:
        Datetime object to include.
    """

    attribute_type: str = 'signing_time'

    def __init__(self, timestamp: datetime):
        self.timestamp = timestamp

    async def build_attr_value(self, dry_run=False) -> cms.Time:
        return cms.Time({'utc_time': core.UTCTime(self.timestamp)})


# TODO: would be nice to also provide an implementation that does this with
#  a "live" revocation checker, for use in the lower-level APIs
# This one relies on the revinfo being available from earlier preprocessing
# steps, which is good enough for internal use.


class AdobeRevinfoProvider(CMSAttributeProvider):
    """
    Yield Adobe-style revocation information for inclusion into a CMS
    object.

    :param value:
        A (pre-formatted) RevocationInfoArchival object.
    """

    attribute_type: str = 'adobe_revocation_info_archival'

    def __init__(self, value: asn1_pdf.RevocationInfoArchival):
        self.value = value

    async def build_attr_value(
        self, dry_run=False
    ) -> Optional[asn1_pdf.RevocationInfoArchival]:
        return self.value


class CMSAlgorithmProtectionProvider(CMSAttributeProvider):
    attribute_type: str = 'cms_algorithm_protection'

    def __init__(
        self, digest_algo: str, signature_algo: algos.SignedDigestAlgorithm
    ):
        self.digest_algo = digest_algo
        self.signature_algo = signature_algo

    async def build_attr_value(
        self, dry_run=False
    ) -> cms.CMSAlgorithmProtection:
        digest_algorithm_args = {'algorithm': self.digest_algo}
        if self.digest_algo == 'shake256':
            # RFC 8419 requirement
            mech = self.signature_algo
            if mech['algorithm'].native == 'ed448':
                digest_algorithm_args = {
                    'algorithm': 'shake256_len',
                    'parameters': core.Integer(512),
                }
        return cms.CMSAlgorithmProtection(
            {
                'digest_algorithm': algos.DigestAlgorithm(
                    digest_algorithm_args
                ),
                'signature_algorithm': self.signature_algo,
            }
        )


class TSTProvider(CMSAttributeProvider):
    def __init__(
        self,
        digest_algorithm: str,
        data_to_ts: bytes,
        timestamper: TimeStamper,
        attr_type: str = 'signature_time_stamp_token',
        prehashed=False,
    ):
        self.attribute_type = attr_type
        self.digest_algorithm = digest_algorithm
        self.timestamper = timestamper
        self.data = data_to_ts
        self.prehashed = prehashed

    async def build_attr_value(self, dry_run=False) -> cms.ContentInfo:
        digest_algorithm = self.digest_algorithm
        if self.prehashed:
            digest = self.data
        else:
            md_spec = get_pyca_cryptography_hash(digest_algorithm)
            md = hashes.Hash(md_spec)
            md.update(self.data)
            digest = md.finalize()
        if dry_run:
            ts_coro = self.timestamper.async_dummy_response(digest_algorithm)
        else:
            ts_coro = self.timestamper.async_timestamp(digest, digest_algorithm)
        return await ts_coro


class SignedAttributeProviderSpec(abc.ABC):
    """
    .. versionadded:: 0.14.0

    Interface for setting up signed attributes, independently of the
    :class:`~pyhanko.sign.signers.pdf_cms.Signer` hierarchy.
    """

    def signed_attr_providers(
        self, data_digest: bytes, digest_algorithm: str
    ) -> Iterable[CMSAttributeProvider]:
        """
        Lazily set up signed attribute providers.

        :param data_digest:
            The digest of the data to be signed.
        :param digest_algorithm:
            The digest algorithm used.
        """
        raise NotImplementedError


class UnsignedAttributeProviderSpec(abc.ABC):
    """
    .. versionadded:: 0.14.0

    Interface for setting up unsigned attributes, independently of the
    :class:`~pyhanko.sign.signers.pdf_cms.Signer` hierarchy.
    """

    def unsigned_attr_providers(
        self,
        signature: bytes,
        signed_attrs: cms.CMSAttributes,
        digest_algorithm: str,
    ) -> Iterable[CMSAttributeProvider]:
        """
        Lazily set up unsigned attribute providers.

        :param signature:
            The signature computed over the signed attributes.
        :param signed_attrs:
            Signed attributes over which the signature was taken.
        :param digest_algorithm:
            The digest algorithm used.
        """
        raise NotImplementedError



import binascii
from datetime import datetime, timedelta
from io import BytesIO

import pytest
import tzlocal
from freezegun import freeze_time
from pyhanko.pdf_utils import crypt, embed, generic, misc, writer
from pyhanko.pdf_utils.crypt import AuthStatus
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader

from .samples import *


def _embed_test(w, fname, ufname, data, created=None, modified=None):
    ef_obj = embed.EmbeddedFileObject.from_file_data(
        w,
        data=data,
        mime_type='application/pdf',
        params=embed.EmbeddedFileParams(
            creation_date=created, modification_date=modified
        ),
    )

    spec = embed.FileSpec(
        file_spec_string=fname,
        file_name=ufname,
        embedded_data=ef_obj,
        description='Embedding test',
    )
    embed.embed_file(w, spec)


@freeze_time('2020-11-01')
@pytest.mark.parametrize('incremental', [True, False])
def test_simple_embed(incremental):
    if incremental:
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    else:
        r = PdfFileReader(BytesIO(MINIMAL))
        w = writer.copy_into_new_writer(r)

    modified = datetime.now(tz=tzlocal.get_localzone())
    created = modified - timedelta(days=1)
    _embed_test(
        w,
        fname='vector-test.pdf',
        ufname='テスト.pdf',
        data=VECTOR_IMAGE_PDF,
        created=created,
        modified=modified,
    )

    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    assert r.input_version == (1, 7)
    emb_lst = r.root['/Names']['/EmbeddedFiles']['/Names']
    assert len(emb_lst) == 2
    assert emb_lst[0] == 'vector-test.pdf'
    spec_obj = emb_lst[1]
    assert spec_obj['/Desc'] == 'Embedding test'
    assert spec_obj['/UF'] == 'テスト.pdf'
    stream = spec_obj['/EF']['/F']
    assert stream.data == VECTOR_IMAGE_PDF

    assert stream['/Subtype'] == '/application/pdf'

    assert stream['/Params']['/CheckSum'] == binascii.unhexlify(
        'caaf24354fd2e68c08826d65b309b404'
    )
    assert generic.parse_pdf_date(stream['/Params']['/ModDate']) == modified
    assert generic.parse_pdf_date(stream['/Params']['/CreationDate']) == created

    assert '/AF' not in r.root


@freeze_time('2020-11-01')
@pytest.mark.parametrize('incremental', [True, False])
def test_embed_twice(incremental):
    r = PdfFileReader(BytesIO(MINIMAL))
    w = writer.copy_into_new_writer(r)

    modified = datetime.now(tz=tzlocal.get_localzone())
    created = modified - timedelta(days=1)
    _embed_test(
        w,
        fname='vector-test.pdf',
        ufname='テスト.pdf',
        data=VECTOR_IMAGE_PDF,
        created=created,
        modified=modified,
    )

    if incremental:
        out = BytesIO()
        w.write(out)
        w = IncrementalPdfFileWriter(out)

    _embed_test(
        w,
        fname='some-other-file.pdf',
        ufname='テスト2.pdf',
        data=MINIMAL_AES256,
        created=created,
        modified=modified,
    )

    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    emb_lst = r.root['/Names']['/EmbeddedFiles']['/Names']
    assert len(emb_lst) == 4
    assert emb_lst[0] == 'vector-test.pdf'
    spec_obj = emb_lst[1]
    assert spec_obj['/UF'] == 'テスト.pdf'
    stream = spec_obj['/EF']['/F']
    assert stream.data == VECTOR_IMAGE_PDF

    assert emb_lst[2] == 'some-other-file.pdf'
    spec_obj = emb_lst[3]
    assert spec_obj['/UF'] == 'テスト2.pdf'
    stream = spec_obj['/EF']['/F']
    assert stream.data == MINIMAL_AES256


@freeze_time('2020-11-01')
@pytest.mark.parametrize('incremental', [True, False])
def test_embed_with_af(incremental):
    if incremental:
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    else:
        r = PdfFileReader(BytesIO(MINIMAL))
        w = writer.copy_into_new_writer(r)

    modified = datetime.now(tz=tzlocal.get_localzone())
    created = modified - timedelta(days=1)
    ef_obj = embed.EmbeddedFileObject.from_file_data(
        w,
        data=VECTOR_IMAGE_PDF,
        params=embed.EmbeddedFileParams(
            creation_date=created, modification_date=modified
        ),
    )

    spec = embed.FileSpec(
        file_spec_string='vector-test.pdf',
        embedded_data=ef_obj,
        description='Embedding test /w assoc file',
        af_relationship=generic.pdf_name('/Unspecified'),
    )
    embed.embed_file(w, spec)
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    assert r.input_version == (2, 0)
    emb_lst = r.root['/Names']['/EmbeddedFiles']['/Names']
    assert len(emb_lst) == 2
    assert emb_lst[0] == 'vector-test.pdf'
    spec_obj = emb_lst[1]
    assert '/UF' not in spec_obj
    assert spec_obj['/AFRelationship'] == '/Unspecified'
    stream = spec_obj['/EF']['/F']
    assert stream.data == VECTOR_IMAGE_PDF
    assert '/UF' not in spec_obj['/EF']

    assert r.root['/AF'].raw_get(0).reference == spec_obj.container_ref


def test_embed_without_ef_stream():
    r = PdfFileReader(BytesIO(MINIMAL))
    w = writer.copy_into_new_writer(r)

    spec = embed.FileSpec(
        file_spec_string='vector-test.pdf',
        description='Embedding test /w assoc file',
        af_relationship=generic.pdf_name('/Unspecified'),
    )
    err_msg = "File spec does not have an embedded file stream"
    with pytest.raises(misc.PdfWriteError, match=err_msg):
        embed.embed_file(w, spec)


def test_encrypt_efs():
    r = PdfFileReader(BytesIO(MINIMAL))
    w = writer.copy_into_new_writer(r)
    cf = crypt.StandardAESCryptFilter(keylen=32)
    cf.set_embedded_only()
    sh = crypt.StandardSecurityHandler.build_from_pw(
        'secret',
        crypt_filter_config=crypt.CryptFilterConfiguration(
            {crypt.STD_CF: cf},
            default_stream_filter=crypt.IDENTITY,
            default_string_filter=crypt.IDENTITY,
            default_file_filter=crypt.STD_CF,
        ),
        encrypt_metadata=False,
    )
    w._assign_security_handler(sh)
    modified = datetime.now(tz=tzlocal.get_localzone())
    created = modified - timedelta(days=1)
    _embed_test(
        w,
        fname='vector-test.pdf',
        ufname='テスト.pdf',
        data=VECTOR_IMAGE_PDF,
        created=created,
        modified=modified,
    )

    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    # should be able to access this without authenticating
    assert b'Hello' in r.root['/Pages']['/Kids'][0]['/Contents'].data
    ef_stm = r.root['/Names']['/EmbeddedFiles']['/Names'][1]['/EF'].raw_get(
        '/F'
    )

    result = r.decrypt('secret')
    assert result.status == AuthStatus.OWNER

    assert ef_stm.get_object()._has_crypt_filter
    assert ef_stm.get_object().data == VECTOR_IMAGE_PDF


def test_decrypt_ef_without_explicit_crypt_filter():
    # such files violate the spec, but since we can deal with them gracefully,
    # we certainly should

    with open(PDF_DATA_DIR + '/embedded-encrypted-nocf.pdf', 'rb') as inf:
        r = PdfFileReader(inf)
        ef_stm = r.root['/Names']['/EmbeddedFiles']['/Names'][1]['/EF'].raw_get(
            '/F'
        )
        r.decrypt('secret')
        assert not ef_stm.get_object()._has_crypt_filter
        assert ef_stm.get_object().data == VECTOR_IMAGE_PDF


def test_wrapper_doc_underspecified():
    with pytest.raises(ValueError, match='exactly one of.*must be'):
        embed.wrap_encrypted_payload(b'bacadsflkj')

    with pytest.raises(ValueError, match='exactly one of.*must be'):
        embed.wrap_encrypted_payload(b'bacadsflkj', password='a', certs=[])


def test_wrapper_doc():
    w = embed.wrap_encrypted_payload(VECTOR_IMAGE_PDF, password='secret')
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    assert b'attached file' in r.root['/Pages']['/Kids'][0]['/Contents'].data

    assert r.root['/Collection']['/D'] == 'attachment.pdf'
    assert r.root['/Collection']['/View'] == '/H'
    ef_stm = r.root['/Names']['/EmbeddedFiles']['/Names'][1]['/EF'].raw_get(
        '/F'
    )

    result = r.decrypt('secret')
    assert result.status == AuthStatus.OWNER

    assert ef_stm.get_object()._has_crypt_filter
    assert ef_stm.get_object().data == VECTOR_IMAGE_PDF


def test_wrapper_doc_pubkey():
    w = embed.wrap_encrypted_payload(
        VECTOR_IMAGE_PDF, certs=[PUBKEY_TEST_DECRYPTER.cert]
    )
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    assert b'attached file' in r.root['/Pages']['/Kids'][0]['/Contents'].data

    assert r.root['/Collection']['/D'] == 'attachment.pdf'
    assert r.root['/Collection']['/View'] == '/H'
    ef_stm = r.root['/Names']['/EmbeddedFiles']['/Names'][1]['/EF'].raw_get(
        '/F'
    )

    result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    assert result.status == AuthStatus.USER

    assert ef_stm.get_object()._has_crypt_filter
    assert ef_stm.get_object().data == VECTOR_IMAGE_PDF
